@@ -113,10 +113,10 @@ init_quarks (void)
 	"    (n %m)"     \
 	"    (e %m)))"
 
-GkrParseResult
+GkrPkixResult
 gkr_pkix_der_read_public_key_rsa (const guchar *data, gsize n_data, gcry_sexp_t *s_key)
 {
-	GkrParseResult ret = GKR_PARSE_UNRECOGNIZED;
+	GkrPkixResult ret = GKR_PKIX_UNRECOGNIZED;
 	ASN1_TYPE asn = ASN1_TYPE_EMPTY;
 	gcry_mpi_t n, e;
 	int res;
@@ -127,7 +127,7 @@ gkr_pkix_der_read_public_key_rsa (const guchar *data, gsize n_data, gcry_sexp_t
 	if (!asn)
 		goto done;
 		
-	ret = GKR_PARSE_FAILURE;
+	ret = GKR_PKIX_FAILURE;
     
 	if (!gkr_pkix_asn1_read_mpi (asn, "modulus", &n) || 
 	    !gkr_pkix_asn1_read_mpi (asn, "publicExponent", &e))
@@ -138,7 +138,7 @@ gkr_pkix_der_read_public_key_rsa (const guchar *data, gsize n_data, gcry_sexp_t
 		goto done;
 
 	g_assert (*s_key);
-	ret = GKR_PARSE_SUCCESS;
+	ret = GKR_PKIX_SUCCESS;
 
 done:
 	if (asn)
@@ -146,7 +146,7 @@ done:
 	gcry_mpi_release (n);
 	gcry_mpi_release (e);
 	
-	if (ret == GKR_PARSE_FAILURE)
+	if (ret == GKR_PKIX_FAILURE)
 		g_message ("invalid RSA public key");
 		
 	return ret;
@@ -162,12 +162,13 @@ done:
 	"    (q %m)"     \
 	"    (u %m)))"
 
-GkrParseResult
+GkrPkixResult
 gkr_pkix_der_read_private_key_rsa (const guchar *data, gsize n_data, gcry_sexp_t *s_key)
 {
-	GkrParseResult ret = GKR_PARSE_UNRECOGNIZED;
+	GkrPkixResult ret = GKR_PKIX_UNRECOGNIZED;
 	gcry_mpi_t n, e, d, p, q, u;
 	gcry_mpi_t tmp;
+	guint version;
 	int res;
 	ASN1_TYPE asn = ASN1_TYPE_EMPTY;
 
@@ -177,7 +178,17 @@ gkr_pkix_der_read_private_key_rsa (const guchar *data, gsize n_data, gcry_sexp_t
 	if (!asn)
 		goto done;
 		
-	ret = GKR_PARSE_FAILURE;
+	ret = GKR_PKIX_FAILURE;
+	
+	if (!gkr_pkix_asn1_read_uint (asn, "version", &version))
+		goto done;
+	
+	/* We only support simple version */
+	if (version != 0) {
+		ret = GKR_PKIX_UNRECOGNIZED;
+		g_message ("unsupported version of RSA key: %u", version);
+		goto done;
+	}
     
 	if (!gkr_pkix_asn1_read_mpi (asn, "modulus", &n) || 
 	    !gkr_pkix_asn1_read_mpi (asn, "publicExponent", &e) ||
@@ -203,7 +214,7 @@ gkr_pkix_der_read_private_key_rsa (const guchar *data, gsize n_data, gcry_sexp_t
 		goto done;
 
 	g_assert (*s_key);
-	ret = GKR_PARSE_SUCCESS;
+	ret = GKR_PKIX_SUCCESS;
 
 done:
 	if (asn)
@@ -215,7 +226,7 @@ done:
 	gcry_mpi_release (q);
 	gcry_mpi_release (u);
 	
-	if (ret == GKR_PARSE_FAILURE)
+	if (ret == GKR_PKIX_FAILURE)
 		g_message ("invalid RSA key");
 		
 	return ret;
@@ -229,10 +240,10 @@ done:
 	"    (g %m)"     \
 	"    (y %m)))"
 
-GkrParseResult
+GkrPkixResult
 gkr_pkix_der_read_public_key_dsa (const guchar *data, gsize n_data, gcry_sexp_t *s_key)
 {
-	GkrParseResult ret = GKR_PARSE_UNRECOGNIZED;
+	GkrPkixResult ret = GKR_PKIX_UNRECOGNIZED;
 	ASN1_TYPE asn = ASN1_TYPE_EMPTY;
 	gcry_mpi_t p, q, g, y;
 	int res;
@@ -243,7 +254,7 @@ gkr_pkix_der_read_public_key_dsa (const guchar *data, gsize n_data, gcry_sexp_t
 	if (!asn)
 		goto done;
 	
-	ret = GKR_PARSE_FAILURE;
+	ret = GKR_PKIX_FAILURE;
     
 	if (!gkr_pkix_asn1_read_mpi (asn, "p", &p) || 
 	    !gkr_pkix_asn1_read_mpi (asn, "q", &q) ||
@@ -256,7 +267,7 @@ gkr_pkix_der_read_public_key_dsa (const guchar *data, gsize n_data, gcry_sexp_t
 		goto done;
 		
 	g_assert (*s_key);
-	ret = GKR_PARSE_SUCCESS;
+	ret = GKR_PKIX_SUCCESS;
 	
 done:
 	if (asn)
@@ -266,19 +277,19 @@ done:
 	gcry_mpi_release (g);
 	gcry_mpi_release (y);
 	
-	if (ret == GKR_PARSE_FAILURE) 
+	if (ret == GKR_PKIX_FAILURE) 
 		g_message ("invalid public DSA key");
 		
 	return ret;	
 }
 
-GkrParseResult
+GkrPkixResult
 gkr_pkix_der_read_public_key_dsa_parts (const guchar *keydata, gsize n_keydata,
                                         const guchar *params, gsize n_params,
                                         gcry_sexp_t *s_key)
 {
 	gcry_mpi_t p, q, g, y;
-	GkrParseResult ret = GKR_PARSE_UNRECOGNIZED;
+	GkrPkixResult ret = GKR_PKIX_UNRECOGNIZED;
 	ASN1_TYPE asn_params = ASN1_TYPE_EMPTY;
 	ASN1_TYPE asn_key = ASN1_TYPE_EMPTY;
 	int res;
@@ -290,7 +301,7 @@ gkr_pkix_der_read_public_key_dsa_parts (const guchar *keydata, gsize n_keydata,
 	if (!asn_params || !asn_key)
 		goto done;
 	
-	ret = GKR_PARSE_FAILURE;
+	ret = GKR_PKIX_FAILURE;
     
 	if (!gkr_pkix_asn1_read_mpi (asn_params, "p", &p) || 
 	    !gkr_pkix_asn1_read_mpi (asn_params, "q", &q) ||
@@ -305,7 +316,7 @@ gkr_pkix_der_read_public_key_dsa_parts (const guchar *keydata, gsize n_keydata,
 		goto done;
 		
 	g_assert (*s_key);
-	ret = GKR_PARSE_SUCCESS;
+	ret = GKR_PKIX_SUCCESS;
 	
 done:
 	if (asn_key)
@@ -317,7 +328,7 @@ done:
 	gcry_mpi_release (g);
 	gcry_mpi_release (y);
 	
-	if (ret == GKR_PARSE_FAILURE) 
+	if (ret == GKR_PKIX_FAILURE) 
 		g_message ("invalid DSA key");
 		
 	return ret;	
@@ -333,11 +344,11 @@ done:
 	"    (y %m)"     \
 	"    (x %m)))"
 
-GkrParseResult
+GkrPkixResult
 gkr_pkix_der_read_private_key_dsa (const guchar *data, gsize n_data, gcry_sexp_t *s_key)
 {
 	gcry_mpi_t p, q, g, y, x;
-	GkrParseResult ret = GKR_PARSE_UNRECOGNIZED;
+	GkrPkixResult ret = GKR_PKIX_UNRECOGNIZED;
 	int res;
 	ASN1_TYPE asn;
 
@@ -347,7 +358,7 @@ gkr_pkix_der_read_private_key_dsa (const guchar *data, gsize n_data, gcry_sexp_t
 	if (!asn)
 		goto done;
 	
-	ret = GKR_PARSE_FAILURE;
+	ret = GKR_PKIX_FAILURE;
     
 	if (!gkr_pkix_asn1_read_mpi (asn, "p", &p) || 
 	    !gkr_pkix_asn1_read_mpi (asn, "q", &q) ||
@@ -361,7 +372,7 @@ gkr_pkix_der_read_private_key_dsa (const guchar *data, gsize n_data, gcry_sexp_t
 		goto done;
 		
 	g_assert (*s_key);
-	ret = GKR_PARSE_SUCCESS;
+	ret = GKR_PKIX_SUCCESS;
 
 done:
 	if (asn)
@@ -372,19 +383,19 @@ done:
 	gcry_mpi_release (y);
 	gcry_mpi_release (x);
 	
-	if (ret == GKR_PARSE_FAILURE) 
+	if (ret == GKR_PKIX_FAILURE) 
 		g_message ("invalid DSA key");
 		
 	return ret;
 }
 
-GkrParseResult
+GkrPkixResult
 gkr_pkix_der_read_private_key_dsa_parts (const guchar *keydata, gsize n_keydata,
                                          const guchar *params, gsize n_params, 
                                          gcry_sexp_t *s_key)
 {
 	gcry_mpi_t p, q, g, y, x;
-	GkrParseResult ret = GKR_PARSE_UNRECOGNIZED;
+	GkrPkixResult ret = GKR_PKIX_UNRECOGNIZED;
 	int res;
 	ASN1_TYPE asn_params = ASN1_TYPE_EMPTY;
 	ASN1_TYPE asn_key = ASN1_TYPE_EMPTY;
@@ -396,7 +407,7 @@ gkr_pkix_der_read_private_key_dsa_parts (const guchar *keydata, gsize n_keydata,
 	if (!asn_params || !asn_key)
 		goto done;
 	
-	ret = GKR_PARSE_FAILURE;
+	ret = GKR_PKIX_FAILURE;
     
 	if (!gkr_pkix_asn1_read_mpi (asn_params, "p", &p) || 
 	    !gkr_pkix_asn1_read_mpi (asn_params, "q", &q) ||
@@ -415,7 +426,7 @@ gkr_pkix_der_read_private_key_dsa_parts (const guchar *keydata, gsize n_keydata,
 		goto done;
 		
 	g_assert (*s_key);
-	ret = GKR_PARSE_SUCCESS;
+	ret = GKR_PKIX_SUCCESS;
 	
 done:
 	if (asn_key)
@@ -428,28 +439,28 @@ done:
 	gcry_mpi_release (y);
 	gcry_mpi_release (x);
 	
-	if (ret == GKR_PARSE_FAILURE) 
+	if (ret == GKR_PKIX_FAILURE) 
 		g_message ("invalid DSA key");
 		
 	return ret;	
 }
 
-GkrParseResult  
+GkrPkixResult  
 gkr_pkix_der_read_public_key (const guchar *data, gsize n_data, gcry_sexp_t *s_key)
 {
-	GkrParseResult res;
+	GkrPkixResult res;
 	
 	res = gkr_pkix_der_read_public_key_rsa (data, n_data, s_key);
-	if (res == GKR_PARSE_UNRECOGNIZED)
+	if (res == GKR_PKIX_UNRECOGNIZED)
 		res = gkr_pkix_der_read_public_key_dsa (data, n_data, s_key);
 		
 	return res;
 }
 
-GkrParseResult
+GkrPkixResult
 gkr_pkix_der_read_public_key_info (const guchar* data, gsize n_data, gcry_sexp_t* s_key)
 {
-	GkrParseResult ret = GKR_PARSE_UNRECOGNIZED;
+	GkrPkixResult ret = GKR_PKIX_UNRECOGNIZED;
 	GQuark oid;
 	ASN1_TYPE asn;
 	gsize n_key, n_params;
@@ -462,10 +473,10 @@ gkr_pkix_der_read_public_key_info (const guchar* data, gsize n_data, gcry_sexp_t
 	if (!asn)
 		goto done;
 	
-	ret = GKR_PARSE_FAILURE;
+	ret = GKR_PKIX_FAILURE;
     
 	/* Figure out the algorithm */
-	oid = gkr_pkix_asn1_read_quark (asn, "algorithm.algorithm");
+	oid = gkr_pkix_asn1_read_oid (asn, "algorithm.algorithm");
 	if (!oid)
 		goto done;
 		
@@ -497,19 +508,19 @@ done:
 	
 	g_free (key);
 		
-	if (ret == GKR_PARSE_FAILURE)
+	if (ret == GKR_PKIX_FAILURE)
 		g_message ("invalid subject public-key info");
 		
 	return ret;
 }
 
-GkrParseResult
+GkrPkixResult
 gkr_pkix_der_read_private_key (const guchar *data, gsize n_data, gcry_sexp_t *s_key)
 {
-	GkrParseResult res;
+	GkrPkixResult res;
 	
 	res = gkr_pkix_der_read_private_key_rsa (data, n_data, s_key);
-	if (res == GKR_PARSE_UNRECOGNIZED)
+	if (res == GKR_PKIX_UNRECOGNIZED)
 		res = gkr_pkix_der_read_private_key_dsa (data, n_data, s_key);
 		
 	return res;
@@ -548,6 +559,73 @@ done:
 	return result;
 }
 
+guchar*
+gkr_pkix_der_write_private_key_rsa (gcry_sexp_t s_key, gsize *n_key)
+{
+	ASN1_TYPE asn = ASN1_TYPE_EMPTY;
+	gcry_mpi_t n, e, d, p, q, u, e1, e2, tmp;
+	guchar *result = NULL;
+	int res;
+
+	n = e = d = p = q = u = e1 = e2 = tmp = NULL;
+
+	res = asn1_create_element (gkr_pkix_asn1_get_pk_asn1type (), 
+	                           "PK.RSAPrivateKey", &asn);
+	g_return_val_if_fail (res == ASN1_SUCCESS, NULL);
+
+	if (!gkr_crypto_sexp_extract_mpi (s_key, &n, "rsa", "n", NULL) || 
+	    !gkr_crypto_sexp_extract_mpi (s_key, &e, "rsa", "e", NULL) ||
+	    !gkr_crypto_sexp_extract_mpi (s_key, &d, "rsa", "d", NULL) ||
+	    !gkr_crypto_sexp_extract_mpi (s_key, &p, "rsa", "p", NULL) ||
+	    !gkr_crypto_sexp_extract_mpi (s_key, &q, "rsa", "q", NULL) ||
+	    !gkr_crypto_sexp_extract_mpi (s_key, &u, "rsa", "u", NULL))
+		goto done;
+	
+	if (!gkr_pkix_asn1_write_mpi (asn, "modulus", n) ||
+	    !gkr_pkix_asn1_write_mpi (asn, "publicExponent", e) || 
+	    !gkr_pkix_asn1_write_mpi (asn, "privateExponent", d) ||
+	    !gkr_pkix_asn1_write_mpi (asn, "prime1", p) ||
+	    !gkr_pkix_asn1_write_mpi (asn, "prime2", q) ||
+	    !gkr_pkix_asn1_write_mpi (asn, "coefficient", u))
+		goto done;
+
+	/* Calculate e1 and e2 */
+	tmp = gcry_mpi_snew (1024);
+	gcry_mpi_sub_ui (tmp, p, 1);
+	e1 = gcry_mpi_snew (1024);
+	gcry_mpi_mod (e1, d, tmp);
+	gcry_mpi_sub_ui (tmp, q, 1);
+	e2 = gcry_mpi_snew (1024);
+	gcry_mpi_mod (e2, d, tmp);
+	
+	/* Write out calculated */
+	if (!gkr_pkix_asn1_write_mpi (asn, "exponent1", e1) ||
+	    !gkr_pkix_asn1_write_mpi (asn, "exponent2", e2))
+		goto done;
+
+	/* Write out the version */
+	if (!gkr_pkix_asn1_write_uint (asn, "version", 0))
+		goto done;
+
+	result = gkr_pkix_asn1_encode (asn, "", n_key, NULL);
+	
+done:
+	if (asn)
+		asn1_delete_structure (&asn);
+	gcry_mpi_release (n);
+	gcry_mpi_release (e);
+	gcry_mpi_release (d);
+	gcry_mpi_release (p);
+	gcry_mpi_release (q);
+	gcry_mpi_release (u);
+	
+	gcry_mpi_release (tmp);
+	gcry_mpi_release (e1);
+	gcry_mpi_release (e2);
+	
+	return result;
+}
+
 guchar*
 gkr_pkix_der_write_public_key_dsa (gcry_sexp_t s_key, gsize *len)
 {
@@ -590,6 +668,117 @@ done:
 	return result;
 }
 
+guchar*
+gkr_pkix_der_write_private_key_dsa_part (gcry_sexp_t skey, gsize *n_key)
+{
+	ASN1_TYPE asn = ASN1_TYPE_EMPTY;
+	gcry_mpi_t x;
+	guchar *result = NULL;
+	int res;
+
+	x = NULL;
+
+	res = asn1_create_element (gkr_pkix_asn1_get_pk_asn1type (), 
+	                           "PK.DSAPrivatePart", &asn);
+	g_return_val_if_fail (res == ASN1_SUCCESS, NULL);
+
+	if (!gkr_crypto_sexp_extract_mpi (skey, &x, "dsa", "x", NULL))
+	    	goto done;
+	
+	if (!gkr_pkix_asn1_write_mpi (asn, "", x))
+	    	goto done;
+
+	result = gkr_pkix_asn1_encode (asn, "", n_key, NULL);
+	
+done:
+	if (asn)
+		asn1_delete_structure (&asn);
+	gcry_mpi_release (x);
+	
+	return result;		
+}
+
+guchar*
+gkr_pkix_der_write_private_key_dsa_params (gcry_sexp_t skey, gsize *n_params)
+{
+	ASN1_TYPE asn = ASN1_TYPE_EMPTY;
+	gcry_mpi_t p, q, g;
+	guchar *result = NULL;
+	int res;
+
+	p = q = g = NULL;
+
+	res = asn1_create_element (gkr_pkix_asn1_get_pk_asn1type (), 
+	                           "PK.DSAParameters", &asn);
+	g_return_val_if_fail (res == ASN1_SUCCESS, NULL);
+
+	if (!gkr_crypto_sexp_extract_mpi (skey, &p, "dsa", "p", NULL) || 
+	    !gkr_crypto_sexp_extract_mpi (skey, &q, "dsa", "q", NULL) ||
+	    !gkr_crypto_sexp_extract_mpi (skey, &g, "dsa", "g", NULL))
+	    	goto done;
+	
+	if (!gkr_pkix_asn1_write_mpi (asn, "p", p) ||
+	    !gkr_pkix_asn1_write_mpi (asn, "q", q) ||
+	    !gkr_pkix_asn1_write_mpi (asn, "g", g))
+	    	goto done;
+
+	result = gkr_pkix_asn1_encode (asn, "", n_params, NULL);
+	
+done:
+	if (asn)
+		asn1_delete_structure (&asn);
+	gcry_mpi_release (p);
+	gcry_mpi_release (q);
+	gcry_mpi_release (g);
+	
+	return result;
+}
+
+guchar*
+gkr_pkix_der_write_private_key_dsa (gcry_sexp_t s_key, gsize *len)
+{
+	ASN1_TYPE asn = ASN1_TYPE_EMPTY;
+	gcry_mpi_t p, q, g, y, x;
+	guchar *result = NULL;
+	int res;
+
+	p = q = g = y = x = NULL;
+
+	res = asn1_create_element (gkr_pkix_asn1_get_pk_asn1type (), 
+	                           "PK.DSAPrivateKey", &asn);
+	g_return_val_if_fail (res == ASN1_SUCCESS, NULL);
+
+	if (!gkr_crypto_sexp_extract_mpi (s_key, &p, "dsa", "p", NULL) || 
+	    !gkr_crypto_sexp_extract_mpi (s_key, &q, "dsa", "q", NULL) ||
+	    !gkr_crypto_sexp_extract_mpi (s_key, &g, "dsa", "g", NULL) ||
+	    !gkr_crypto_sexp_extract_mpi (s_key, &y, "dsa", "y", NULL) ||
+	    !gkr_crypto_sexp_extract_mpi (s_key, &x, "dsa", "x", NULL))
+	    	goto done;
+	
+	if (!gkr_pkix_asn1_write_mpi (asn, "p", p) ||
+	    !gkr_pkix_asn1_write_mpi (asn, "q", q) ||
+	    !gkr_pkix_asn1_write_mpi (asn, "g", g) ||
+	    !gkr_pkix_asn1_write_mpi (asn, "Y", y) ||
+	    !gkr_pkix_asn1_write_mpi (asn, "priv", x))
+	    	goto done;
+
+	if (!gkr_pkix_asn1_write_uint (asn, "version", 0))
+		goto done; 
+		
+	result = gkr_pkix_asn1_encode (asn, "", len, NULL);
+	
+done:
+	if (asn)
+		asn1_delete_structure (&asn);
+	gcry_mpi_release (p);
+	gcry_mpi_release (q);
+	gcry_mpi_release (g);
+	gcry_mpi_release (y);
+	gcry_mpi_release (x);
+	
+	return result;
+}
+
 guchar*
 gkr_pkix_der_write_public_key (gcry_sexp_t s_key, gsize *len)
 {
@@ -613,32 +802,55 @@ gkr_pkix_der_write_public_key (gcry_sexp_t s_key, gsize *len)
 	}
 }
 
+guchar*
+gkr_pkix_der_write_private_key (gcry_sexp_t s_key, gsize *len)
+{
+	gboolean is_priv;
+	int algorithm;
+	
+	g_return_val_if_fail (s_key != NULL, NULL);
+	
+	if (!gkr_crypto_skey_parse (s_key, &algorithm, &is_priv, NULL))
+		g_return_val_if_reached (NULL);
+	
+	g_return_val_if_fail (is_priv, NULL);
+		
+	switch (algorithm) {
+	case GCRY_PK_RSA:
+		return gkr_pkix_der_write_private_key_rsa (s_key, len);
+	case GCRY_PK_DSA:
+		return gkr_pkix_der_write_private_key_dsa (s_key, len);
+	default:
+		g_return_val_if_reached (NULL);
+	}
+}
+
 /* -----------------------------------------------------------------------------
  * CERTIFICATES
  */
  
-GkrParseResult
+GkrPkixResult
 gkr_pkix_der_read_certificate (const guchar *data, gsize n_data, ASN1_TYPE *asn1)
 {
 	*asn1 = gkr_pkix_asn1_decode ("PKIX1.Certificate", data, n_data);
 	if (!*asn1)
-		return GKR_PARSE_UNRECOGNIZED;
+		return GKR_PKIX_UNRECOGNIZED;
 	
-	return GKR_PARSE_SUCCESS;
+	return GKR_PKIX_SUCCESS;
 }
 
-GkrParseResult
+GkrPkixResult
 gkr_pkix_der_read_basic_constraints (const guchar *data, gsize n_data, 
                                      gboolean *is_ca, guint *path_len)
 {
-	GkrParseResult ret = GKR_PARSE_UNRECOGNIZED;
+	GkrPkixResult ret = GKR_PKIX_UNRECOGNIZED;
 	ASN1_TYPE asn;
 
 	asn = gkr_pkix_asn1_decode ("PKIX1.BasicConstraints", data, n_data);
 	if (!asn)
 		goto done;
 	
-	ret = GKR_PARSE_FAILURE;
+	ret = GKR_PKIX_FAILURE;
     
     	if (path_len) {
     		if (!gkr_pkix_asn1_read_uint (asn, "pathLenConstraint", path_len))
@@ -646,102 +858,175 @@ gkr_pkix_der_read_basic_constraints (const guchar *data, gsize n_data,
     	}
     	
     	if (is_ca) {
-    		*is_ca = FALSE;
     		if (!gkr_pkix_asn1_read_boolean (asn, "cA", is_ca))
-    			goto done;
+    			*is_ca = FALSE;
     	}
     	
-	ret = GKR_PARSE_SUCCESS;
+	ret = GKR_PKIX_SUCCESS;
 
 done:
 	if (asn)
 		asn1_delete_structure (&asn);
 	
-	if (ret == GKR_PARSE_FAILURE) 
+	if (ret == GKR_PKIX_FAILURE) 
 		g_message ("invalid basic constraints");
 		
 	return ret;
 }
 
+GkrPkixResult
+gkr_pkix_der_read_key_usage (const guchar *data, gsize n_data, guint *key_usage)
+{
+	GkrPkixResult ret = GKR_PKIX_UNRECOGNIZED;
+	ASN1_TYPE asn;
+	guchar buf[4];
+	int res, len;
+	
+	asn = gkr_pkix_asn1_decode ("PKIX1.KeyUsage", data, n_data);
+	if (!asn)
+		goto done;
+		
+	ret = GKR_PKIX_FAILURE;
+
+	memset (buf, 0, sizeof (buf));
+	len = sizeof (buf);
+  	res = asn1_read_value (asn, "", buf, &len);
+  	if (res != ASN1_SUCCESS)
+  		goto done;
+
+	*key_usage = buf[0] | (buf[1] << 8);
+	ret = GKR_PKIX_SUCCESS;
+	
+done:
+	if (asn)
+		asn1_delete_structure (&asn);		
+	return ret;
+}
+
+GkrPkixResult
+gkr_pkix_der_read_enhanced_usage (const guchar *data, gsize n_data, GQuark **usage_oids)
+{
+	GkrPkixResult ret = GKR_PKIX_UNRECOGNIZED;
+	ASN1_TYPE asn;
+	gchar *part;
+	GArray *array;
+	GQuark oid;
+	int i;
+	
+	asn = gkr_pkix_asn1_decode ("PKIX1.ExtKeyUsageSyntax", data, n_data);
+	if (!asn)
+		goto done;
+		
+	ret = GKR_PKIX_FAILURE;
+	
+	array = g_array_new (TRUE, TRUE, sizeof (GQuark));
+	for (i = 0; TRUE; ++i) {
+		part = g_strdup_printf ("?%d", i + 1);
+		oid = gkr_pkix_asn1_read_oid (asn, part);
+		g_free (part);
+		
+		if (!oid) 
+			break;
+		
+		g_array_append_val (array, oid);
+	}
+	
+	*usage_oids = (GQuark*)g_array_free (array, FALSE);
+	ret = GKR_PKIX_SUCCESS;
+	
+done:
+	if (asn)
+		asn1_delete_structure (&asn);
+	return ret;
+}
+
+guchar*
+gkr_pkix_der_write_certificate (ASN1_TYPE asn1, gsize *n_data)
+{
+	g_return_val_if_fail (asn1, NULL);
+	g_return_val_if_fail (n_data, NULL);
+	
+	return gkr_pkix_asn1_encode (asn1, "", n_data, NULL);
+}
+
 /* -----------------------------------------------------------------------------
  * CIPHER/KEY DESCRIPTIONS 
  */
  
-GkrParseResult
-gkr_pkix_der_read_cipher (GkrPkixParser *parser, GQuark oid_scheme, const gchar *password, 
+GkrPkixResult
+gkr_pkix_der_read_cipher (GQuark oid_scheme, const gchar *password, 
                           const guchar *data, gsize n_data, gcry_cipher_hd_t *cih)
 {
-	GkrParseResult ret = GKR_PARSE_UNRECOGNIZED;
+	GkrPkixResult ret = GKR_PKIX_UNRECOGNIZED;
 	
-	g_return_val_if_fail (GKR_IS_PKIX_PARSER (parser), GKR_PARSE_FAILURE);
-	g_return_val_if_fail (oid_scheme != 0, GKR_PARSE_FAILURE);
-	g_return_val_if_fail (cih != NULL, GKR_PARSE_FAILURE);
-	g_return_val_if_fail (data != NULL && n_data != 0, GKR_PARSE_FAILURE);
-	g_return_val_if_fail (password != NULL, GKR_PARSE_FAILURE);
+	g_return_val_if_fail (oid_scheme != 0, GKR_PKIX_FAILURE);
+	g_return_val_if_fail (cih != NULL, GKR_PKIX_FAILURE);
+	g_return_val_if_fail (data != NULL && n_data != 0, GKR_PKIX_FAILURE);
+	g_return_val_if_fail (password != NULL, GKR_PKIX_FAILURE);
 	
 	init_quarks ();
 	
 	/* PKCS#5 PBE */
 	if (oid_scheme == OID_PBE_MD2_DES_CBC)
-		ret = gkr_pkix_der_read_cipher_pkcs5_pbe (parser, GCRY_CIPHER_DES, GCRY_CIPHER_MODE_CBC,
+		ret = gkr_pkix_der_read_cipher_pkcs5_pbe (GCRY_CIPHER_DES, GCRY_CIPHER_MODE_CBC,
 		                                          GCRY_MD_MD2, password, data, n_data, cih);
 
 	else if (oid_scheme == OID_PBE_MD2_RC2_CBC)
 		/* RC2-64 has no implementation in libgcrypt */
-		ret = GKR_PARSE_UNRECOGNIZED;
+		ret = GKR_PKIX_UNRECOGNIZED;
 	else if (oid_scheme == OID_PBE_MD5_DES_CBC)
-		ret = gkr_pkix_der_read_cipher_pkcs5_pbe (parser, GCRY_CIPHER_DES, GCRY_CIPHER_MODE_CBC,
+		ret = gkr_pkix_der_read_cipher_pkcs5_pbe (GCRY_CIPHER_DES, GCRY_CIPHER_MODE_CBC,
 		                                          GCRY_MD_MD5, password, data, n_data, cih);
 	else if (oid_scheme == OID_PBE_MD5_RC2_CBC)
 		/* RC2-64 has no implementation in libgcrypt */
-		ret = GKR_PARSE_UNRECOGNIZED;
+		ret = GKR_PKIX_UNRECOGNIZED;
 	else if (oid_scheme == OID_PBE_SHA1_DES_CBC)
-		ret = gkr_pkix_der_read_cipher_pkcs5_pbe (parser, GCRY_CIPHER_DES, GCRY_CIPHER_MODE_CBC,
+		ret = gkr_pkix_der_read_cipher_pkcs5_pbe (GCRY_CIPHER_DES, GCRY_CIPHER_MODE_CBC,
 		                                          GCRY_MD_SHA1, password, data, n_data, cih);
 	else if (oid_scheme == OID_PBE_SHA1_RC2_CBC)
 		/* RC2-64 has no implementation in libgcrypt */
-		ret = GKR_PARSE_UNRECOGNIZED;
+		ret = GKR_PKIX_UNRECOGNIZED;
 
 	
 	/* PKCS#5 PBES2 */
 	else if (oid_scheme == OID_PBES2)
-		ret = gkr_pkix_der_read_cipher_pkcs5_pbes2 (parser, password, data, n_data, cih);
+		ret = gkr_pkix_der_read_cipher_pkcs5_pbes2 (password, data, n_data, cih);
 
 		
 	/* PKCS#12 PBE */
 	else if (oid_scheme == OID_PKCS12_PBE_ARCFOUR_SHA1)
-		ret = gkr_pkix_der_read_cipher_pkcs12_pbe (parser, GCRY_CIPHER_ARCFOUR, GCRY_CIPHER_MODE_STREAM, 
-	                                                   password, data, n_data, cih);
+		ret = gkr_pkix_der_read_cipher_pkcs12_pbe (GCRY_CIPHER_ARCFOUR, GCRY_CIPHER_MODE_STREAM, 
+                                                   password, data, n_data, cih);
 	else if (oid_scheme == OID_PKCS12_PBE_RC4_40_SHA1)
 		/* RC4-40 has no implementation in libgcrypt */;
 
 	else if (oid_scheme == OID_PKCS12_PBE_3DES_SHA1)
-		ret = gkr_pkix_der_read_cipher_pkcs12_pbe (parser, GCRY_CIPHER_3DES, GCRY_CIPHER_MODE_CBC, 
-	                                                   password, data, n_data, cih);
+		ret = gkr_pkix_der_read_cipher_pkcs12_pbe (GCRY_CIPHER_3DES, GCRY_CIPHER_MODE_CBC, 
+                                                   password, data, n_data, cih);
 	else if (oid_scheme == OID_PKCS12_PBE_2DES_SHA1) 
 		/* 2DES has no implementation in libgcrypt */;
 		
 	else if (oid_scheme == OID_PKCS12_PBE_RC2_128_SHA1)
-		ret = gkr_pkix_der_read_cipher_pkcs12_pbe (parser, GCRY_CIPHER_RFC2268_128, GCRY_CIPHER_MODE_CBC, 
-	                                                   password, data, n_data, cih);
+		ret = gkr_pkix_der_read_cipher_pkcs12_pbe (GCRY_CIPHER_RFC2268_128, GCRY_CIPHER_MODE_CBC, 
+                                                   password, data, n_data, cih);
 
 	else if (oid_scheme == OID_PKCS12_PBE_RC2_40_SHA1)
-		ret = gkr_pkix_der_read_cipher_pkcs12_pbe (parser, GCRY_CIPHER_RFC2268_40, GCRY_CIPHER_MODE_CBC, 
-	                                                   password, data, n_data, cih);
+		ret = gkr_pkix_der_read_cipher_pkcs12_pbe (GCRY_CIPHER_RFC2268_40, GCRY_CIPHER_MODE_CBC, 
+                                                   password, data, n_data, cih);
 
-	if (ret == GKR_PARSE_UNRECOGNIZED)
+	if (ret == GKR_PKIX_UNRECOGNIZED)
     		g_message ("unsupported or unrecognized cipher oid: %s", g_quark_to_string (oid_scheme));
     	return ret;
 }
 
-GkrParseResult
-gkr_pkix_der_read_cipher_pkcs5_pbe (GkrPkixParser *parser, int cipher_algo, int cipher_mode, 
+GkrPkixResult
+gkr_pkix_der_read_cipher_pkcs5_pbe (int cipher_algo, int cipher_mode, 
                                     int hash_algo, const gchar *password, const guchar *data, 
                                     gsize n_data, gcry_cipher_hd_t *cih)
 {
 	ASN1_TYPE asn = ASN1_TYPE_EMPTY;
 	gcry_error_t gcry;
-	GkrParseResult ret;
+	GkrPkixResult ret;
 	const guchar *salt;
 	gsize n_salt;
 	gsize n_block, n_key;
@@ -749,14 +1034,13 @@ gkr_pkix_der_read_cipher_pkcs5_pbe (GkrPkixParser *parser, int cipher_algo, int
 	guchar *key = NULL;
 	guchar *iv = NULL;
 
-	g_return_val_if_fail (GKR_IS_PKIX_PARSER (parser), GKR_PARSE_FAILURE);
-	g_return_val_if_fail (cipher_algo != 0 && cipher_mode != 0, GKR_PARSE_FAILURE);
-	g_return_val_if_fail (cih != NULL, GKR_PARSE_FAILURE);
-	g_return_val_if_fail (data != NULL && n_data != 0, GKR_PARSE_FAILURE);
-	g_return_val_if_fail (password != NULL, GKR_PARSE_FAILURE);
+	g_return_val_if_fail (cipher_algo != 0 && cipher_mode != 0, GKR_PKIX_FAILURE);
+	g_return_val_if_fail (cih != NULL, GKR_PKIX_FAILURE);
+	g_return_val_if_fail (data != NULL && n_data != 0, GKR_PKIX_FAILURE);
+	g_return_val_if_fail (password != NULL, GKR_PKIX_FAILURE);
 
 	*cih = NULL;	
-	ret = GKR_PARSE_UNRECOGNIZED;
+	ret = GKR_PKIX_UNRECOGNIZED;
 	
 	/* Check if we can use this algorithm */
 	if (gcry_cipher_algo_info (cipher_algo, GCRYCTL_TEST_ALGO, NULL, 0) != 0 ||
@@ -767,7 +1051,7 @@ gkr_pkix_der_read_cipher_pkcs5_pbe (GkrPkixParser *parser, int cipher_algo, int
 	if (!asn) 
 		goto done;
 		
-	ret = GKR_PARSE_FAILURE;
+	ret = GKR_PKIX_FAILURE;
 		
 	salt = gkr_pkix_asn1_read_content (asn, data, n_data, "salt", &n_salt);
 	if (!salt)
@@ -776,7 +1060,7 @@ gkr_pkix_der_read_cipher_pkcs5_pbe (GkrPkixParser *parser, int cipher_algo, int
 		iterations = 1;
 		
 	n_key = gcry_cipher_get_algo_keylen (cipher_algo);
-	g_return_val_if_fail (n_key > 0, GKR_PARSE_FAILURE);
+	g_return_val_if_fail (n_key > 0, GKR_PKIX_FAILURE);
 	n_block = gcry_cipher_get_algo_blklen (cipher_algo);
 		
 	if (!gkr_crypto_generate_symkey_pbe (cipher_algo, hash_algo, password, salt,
@@ -793,7 +1077,7 @@ gkr_pkix_der_read_cipher_pkcs5_pbe (GkrPkixParser *parser, int cipher_algo, int
 		gcry_cipher_setiv (*cih, iv, n_block);
 	gcry_cipher_setkey (*cih, key, n_key);
 	
-	ret = GKR_PARSE_SUCCESS;
+	ret = GKR_PKIX_SUCCESS;
 
 done:
 	gkr_secure_free (iv);
@@ -806,8 +1090,7 @@ done:
 }
 
 static gboolean
-setup_pkcs5_rc2_params (GkrPkixParser *parser, const guchar *data, guchar n_data,
-                        gcry_cipher_hd_t cih)
+setup_pkcs5_rc2_params (const guchar *data, guchar n_data, gcry_cipher_hd_t cih)
 {
 	ASN1_TYPE asn = ASN1_TYPE_EMPTY;
 	gcry_error_t gcry;
@@ -819,30 +1102,29 @@ setup_pkcs5_rc2_params (GkrPkixParser *parser, const guchar *data, guchar n_data
 
 	asn = gkr_pkix_asn1_decode ("PKIX1.pkcs-5-rc2-CBC-params", data, n_data);
 	if (!asn) 
-		return GKR_PARSE_UNRECOGNIZED;
+		return GKR_PKIX_UNRECOGNIZED;
 		
 	if (!gkr_pkix_asn1_read_uint (asn, "rc2ParameterVersion", &version))
-		return GKR_PARSE_FAILURE;
+		return GKR_PKIX_FAILURE;
 	
 	iv = gkr_pkix_asn1_read_content (asn, data, n_data, "iv", &n_iv);
 	asn1_delete_structure (&asn);
 
 	if (!iv)
-		return GKR_PARSE_FAILURE;
+		return GKR_PKIX_FAILURE;
 		
 	gcry = gcry_cipher_setiv (cih, iv, n_iv);
 			
 	if (gcry != 0) {
 		g_message ("couldn't set %lu byte iv on cipher", (gulong)n_iv);
-		return GKR_PARSE_FAILURE;
+		return GKR_PKIX_FAILURE;
 	}
 	
-	return GKR_PARSE_SUCCESS;
+	return GKR_PKIX_SUCCESS;
 }
 
 static gboolean
-setup_pkcs5_des_params (GkrPkixParser *parser, const guchar *data, guchar n_data,
-                        gcry_cipher_hd_t cih)
+setup_pkcs5_des_params (const guchar *data, guchar n_data, gcry_cipher_hd_t cih)
 {
 	ASN1_TYPE asn = ASN1_TYPE_EMPTY;
 	gcry_error_t gcry;
@@ -855,30 +1137,30 @@ setup_pkcs5_des_params (GkrPkixParser *parser, const guchar *data, guchar n_data
 	if (!asn)
 		asn = gkr_pkix_asn1_decode ("PKIX1.pkcs-5-des-CBC-params", data, n_data);
 	if (!asn) 
-		return GKR_PARSE_UNRECOGNIZED;
+		return GKR_PKIX_UNRECOGNIZED;
 	
 	iv = gkr_pkix_asn1_read_content (asn, data, n_data, "", &n_iv);
 	asn1_delete_structure (&asn);
 
 	if (!iv)
-		return GKR_PARSE_FAILURE;
+		return GKR_PKIX_FAILURE;
 		
 	gcry = gcry_cipher_setiv (cih, iv, n_iv);
 			
 	if (gcry != 0) {
 		g_message ("couldn't set %lu byte iv on cipher", (gulong)n_iv);
-		return GKR_PARSE_FAILURE;
+		return GKR_PKIX_FAILURE;
 	}
 	
-	return GKR_PARSE_SUCCESS;
+	return GKR_PKIX_SUCCESS;
 }
 
-static GkrParseResult
-setup_pkcs5_pbkdf2_params (GkrPkixParser *parser, const gchar *password, const guchar *data, 
+static GkrPkixResult
+setup_pkcs5_pbkdf2_params (const gchar *password, const guchar *data, 
                            gsize n_data, int cipher_algo, gcry_cipher_hd_t cih)
 {
 	ASN1_TYPE asn = ASN1_TYPE_EMPTY;
-	GkrParseResult ret;
+	GkrPkixResult ret;
 	gcry_error_t gcry;
 	guchar *key = NULL; 
 	const guchar *salt;
@@ -889,13 +1171,13 @@ setup_pkcs5_pbkdf2_params (GkrPkixParser *parser, const gchar *password, const g
 	g_assert (cipher_algo);
 	g_assert (data);
 	
-	ret = GKR_PARSE_UNRECOGNIZED;
+	ret = GKR_PKIX_UNRECOGNIZED;
 
 	asn = gkr_pkix_asn1_decode ("PKIX1.pkcs-5-PBKDF2-params", data, n_data);
 	if (!asn)
 		goto done;
 		
-	ret = GKR_PARSE_FAILURE;
+	ret = GKR_PKIX_FAILURE;
 		
 	if (!gkr_pkix_asn1_read_uint (asn, "iterationCount", &iterations))
 		iterations = 1;
@@ -908,7 +1190,7 @@ setup_pkcs5_pbkdf2_params (GkrPkixParser *parser, const gchar *password, const g
 		goto done;
 
 	n_key = gcry_cipher_get_algo_keylen (cipher_algo);
-	g_return_val_if_fail (n_key > 0, GKR_PARSE_FAILURE);
+	g_return_val_if_fail (n_key > 0, GKR_PKIX_FAILURE);
 	
 	gcry = gcry_cipher_setkey (cih, key, n_key);
 	if (gcry != 0) {
@@ -916,7 +1198,7 @@ setup_pkcs5_pbkdf2_params (GkrPkixParser *parser, const gchar *password, const g
 		goto done;
 	}
 	
-	ret = GKR_PARSE_SUCCESS;
+	ret = GKR_PKIX_SUCCESS;
 	                                         
 done:
 	gkr_secure_free (key);
@@ -925,36 +1207,35 @@ done:
 	return ret;
 }
 
-GkrParseResult
-gkr_pkix_der_read_cipher_pkcs5_pbes2 (GkrPkixParser *parser, const gchar *password, const guchar *data, 
+GkrPkixResult
+gkr_pkix_der_read_cipher_pkcs5_pbes2 (const gchar *password, const guchar *data, 
                                       gsize n_data, gcry_cipher_hd_t *cih)
 {
 	ASN1_TYPE asn = ASN1_TYPE_EMPTY;
-	GkrParseResult r, ret;
+	GkrPkixResult r, ret;
 	GQuark key_deriv_algo, enc_oid;
 	gcry_error_t gcry;
 	int algo, mode;
 	int beg, end, res;
 
-	g_return_val_if_fail (GKR_IS_PKIX_PARSER (parser), GKR_PARSE_FAILURE);
-	g_return_val_if_fail (cih != NULL, GKR_PARSE_FAILURE);
-	g_return_val_if_fail (data != NULL && n_data != 0, GKR_PARSE_FAILURE);
-	g_return_val_if_fail (password != NULL, GKR_PARSE_FAILURE);
+	g_return_val_if_fail (cih != NULL, GKR_PKIX_FAILURE);
+	g_return_val_if_fail (data != NULL && n_data != 0, GKR_PKIX_FAILURE);
+	g_return_val_if_fail (password != NULL, GKR_PKIX_FAILURE);
 	
 	init_quarks ();
 	
 	*cih = NULL;
-	ret = GKR_PARSE_UNRECOGNIZED;
+	ret = GKR_PKIX_UNRECOGNIZED;
 	
 	asn = gkr_pkix_asn1_decode ("PKIX1.pkcs-5-PBES2-params", data, n_data);
 	if (!asn)
 		goto done;
 		
-	res = GKR_PARSE_FAILURE;
+	res = GKR_PKIX_FAILURE;
 	algo = mode = 0;
 	
 	/* Read in all the encryption type */
-	enc_oid = gkr_pkix_asn1_read_quark (asn, "encryptionScheme.algorithm");
+	enc_oid = gkr_pkix_asn1_read_oid (asn, "encryptionScheme.algorithm");
 	if (!enc_oid)
 		goto done;	
 	if (enc_oid == OID_DES_EDE3_CBC)
@@ -968,7 +1249,7 @@ gkr_pkix_der_read_cipher_pkcs5_pbes2 (GkrPkixParser *parser, const gchar *passwo
 	
 	/* Unsupported? */
 	if (algo == 0 || gcry_cipher_algo_info (algo, GCRYCTL_TEST_ALGO, NULL, 0) != 0) {
-		ret = GKR_PARSE_UNRECOGNIZED;
+		ret = GKR_PKIX_UNRECOGNIZED;
 		goto done;
 	}
 
@@ -987,30 +1268,30 @@ gkr_pkix_der_read_cipher_pkcs5_pbes2 (GkrPkixParser *parser, const gchar *passwo
 	switch (algo) {
 	case GCRY_CIPHER_3DES:
 	case GCRY_CIPHER_DES:
-		r = setup_pkcs5_des_params (parser, data + beg, end - beg + 1, *cih);
+		r = setup_pkcs5_des_params (data + beg, end - beg + 1, *cih);
 		break;
 	case GCRY_CIPHER_RFC2268_128:
-		r = setup_pkcs5_rc2_params (parser, data + beg, end - beg + 1, *cih);
+		r = setup_pkcs5_rc2_params (data + beg, end - beg + 1, *cih);
 		break;
 	default:
 		/* Should have been caught on the oid check above */
 		g_assert_not_reached ();
-		r = GKR_PARSE_UNRECOGNIZED;
+		r = GKR_PKIX_UNRECOGNIZED;
 		break;
 	};
 
-	if (r != GKR_PARSE_SUCCESS) {
+	if (r != GKR_PKIX_SUCCESS) {
 		ret = r;
 		goto done;
 	}
 
 	/* Read out the key creation paramaters */
-	key_deriv_algo = gkr_pkix_asn1_read_quark (asn, "keyDerivationFunc.algorithm");
+	key_deriv_algo = gkr_pkix_asn1_read_oid (asn, "keyDerivationFunc.algorithm");
 	if (!key_deriv_algo)
 		goto done;
 	if (key_deriv_algo != OID_PBKDF2) {
 		g_message ("unsupported key derivation algorithm: %s", g_quark_to_string (key_deriv_algo));
-		ret = GKR_PARSE_UNRECOGNIZED;
+		ret = GKR_PKIX_UNRECOGNIZED;
 		goto done;
 	}
 
@@ -1018,11 +1299,10 @@ gkr_pkix_der_read_cipher_pkcs5_pbes2 (GkrPkixParser *parser, const gchar *passwo
 	                                &beg, &end) != ASN1_SUCCESS)
 		goto done;
 	
-	ret = setup_pkcs5_pbkdf2_params (parser, password, data + beg, 
-	                                 end - beg + 1, algo, *cih);
+	ret = setup_pkcs5_pbkdf2_params (password, data + beg, end - beg + 1, algo, *cih);
 
 done:
-	if (ret != GKR_PARSE_SUCCESS && *cih) {
+	if (ret != GKR_PKIX_SUCCESS && *cih) {
 		gcry_cipher_close (*cih);
 		*cih = NULL;
 	}
@@ -1033,14 +1313,13 @@ done:
 	return ret;
 }
 
-GkrParseResult
-gkr_pkix_der_read_cipher_pkcs12_pbe (GkrPkixParser *parser, int cipher_algo, int cipher_mode, 
-                                     const gchar *password, const guchar *data, gsize n_data, 
-                                     gcry_cipher_hd_t *cih)
+GkrPkixResult
+gkr_pkix_der_read_cipher_pkcs12_pbe (int cipher_algo, int cipher_mode, const gchar *password, 
+                                     const guchar *data, gsize n_data, gcry_cipher_hd_t *cih)
 {
 	ASN1_TYPE asn = ASN1_TYPE_EMPTY;
 	gcry_error_t gcry;
-	GkrParseResult ret;
+	GkrPkixResult ret;
 	const guchar *salt;
 	gsize n_salt;
 	gsize n_block, n_key;
@@ -1048,14 +1327,13 @@ gkr_pkix_der_read_cipher_pkcs12_pbe (GkrPkixParser *parser, int cipher_algo, int
 	guchar *key = NULL;
 	guchar *iv = NULL;
 	
-	g_return_val_if_fail (GKR_IS_PKIX_PARSER (parser), GKR_PARSE_FAILURE);
-	g_return_val_if_fail (cipher_algo != 0 && cipher_mode != 0, GKR_PARSE_FAILURE);
-	g_return_val_if_fail (cih != NULL, GKR_PARSE_FAILURE);
-	g_return_val_if_fail (data != NULL && n_data != 0, GKR_PARSE_FAILURE);
-	g_return_val_if_fail (password != NULL, GKR_PARSE_FAILURE);
+	g_return_val_if_fail (cipher_algo != 0 && cipher_mode != 0, GKR_PKIX_FAILURE);
+	g_return_val_if_fail (cih != NULL, GKR_PKIX_FAILURE);
+	g_return_val_if_fail (data != NULL && n_data != 0, GKR_PKIX_FAILURE);
+	g_return_val_if_fail (password != NULL, GKR_PKIX_FAILURE);
 	
 	*cih = NULL;
-	ret = GKR_PARSE_UNRECOGNIZED;
+	ret = GKR_PKIX_UNRECOGNIZED;
 	
 	/* Check if we can use this algorithm */
 	if (gcry_cipher_algo_info (cipher_algo, GCRYCTL_TEST_ALGO, NULL, 0) != 0)
@@ -1065,7 +1343,7 @@ gkr_pkix_der_read_cipher_pkcs12_pbe (GkrPkixParser *parser, int cipher_algo, int
 	if (!asn)
 		goto done;
 
-	ret = GKR_PARSE_FAILURE;
+	ret = GKR_PKIX_FAILURE;
 
 	salt = gkr_pkix_asn1_read_content (asn, data, n_data, "salt", &n_salt);
 	if (!salt)
@@ -1092,10 +1370,10 @@ gkr_pkix_der_read_cipher_pkcs12_pbe (GkrPkixParser *parser, int cipher_algo, int
 		gcry_cipher_setiv (*cih, iv, n_block);
 	gcry_cipher_setkey (*cih, key, n_key);
 	
-	ret = GKR_PARSE_SUCCESS;
+	ret = GKR_PKIX_SUCCESS;
 	
 done:
-	if (ret != GKR_PARSE_SUCCESS && *cih) {
+	if (ret != GKR_PKIX_SUCCESS && *cih) {
 		gcry_cipher_close (*cih);
 		*cih = NULL;
 	}
